@@ -3,25 +3,28 @@ import json
 import pathlib
 import itertools
 import collections
+import urllib.parse
 
 import attr
-from csvw.metadata import TableGroup, Table, Column, Link, Schema
+from csvw.metadata import TableGroup, Table, Column, Link, Schema, is_url
 from csvw import datatypes
 from csvw.dsv import iterrows
 from clldutils.path import git_describe, walk
-from clldutils.misc import log_or_raise
+from clldutils.misc import log_or_raise, lazyproperty
 from clldutils import jsonlib
 
 from pycldf.sources import Sources
-from pycldf.util import pkg_path, resolve_slices
+from pycldf.util import pkg_path, resolve_slices, DictTuple
 from pycldf.terms import term_uri, TERMS, get_column_names, URL as TERMS_URL
 from pycldf.validators import VALIDATORS
+from pycldf import orm
 
 __all__ = [
     'Dataset', 'Generic', 'Wordlist', 'ParallelText', 'Dictionary', 'StructureDataset',
     'iter_datasets']
 
 MD_SUFFIX = '-metadata.json'
+ORM_CLASSES = {cls.component_name(): cls for cls in orm.Object.__subclasses__()}
 
 
 def sniff(p):
@@ -143,6 +146,7 @@ class Dataset(object):
         self.tablegroup = tablegroup
         self.auto_constraints()
         self.sources = Sources.from_file(self.bibpath)
+        self._objects = collections.defaultdict(collections.OrderedDict)
 
     @property
     def metadata_dict(self):
@@ -163,7 +167,7 @@ class Dataset(object):
 
         Note that this property is computed each time it is accessed (because the dataset
         schema may have changed). So when accessing a dataset for reading only, calling code
-        should retrieve `column_names` once, and then work with the local reference.
+        should use `readonly_column_names`.
 
         :return: an `argparse.Namespace` object, with attributes `<object>s` for each component \
         `<Object>Table` defined in the ontology. Each such attribute evaluates to `None` if the \
@@ -173,6 +177,13 @@ class Dataset(object):
         """
         return get_column_names(self)
 
+    @lazyproperty
+    def readonly_column_names(self):
+        """
+        :return: `argparse.Namespace` with component names as attributes.
+        """
+        return get_column_names(self, use_component_names=True, with_multiplicity=True)
+
     def add_provenance(self, **kw):
         """
         Add metadata about the dataset's provenance.
@@ -389,8 +400,17 @@ class Dataset(object):
     @property
     def bibpath(self):
         # Specifying "dc:source": "" means lookup the default location.
+        if is_url(self.directory):
+            return urllib.parse.urljoin(
+                self.directory, self.properties.get('dc:source') or 'sources.bib')
         return self.directory.joinpath(self.properties.get('dc:source') or 'sources.bib')
 
+    @property
+    def bibname(self):
+        if is_url(self.bibpath):
+            return pathlib.Path(urllib.parse.urlparse(self.bibpath).path).name
+        return self.bibpath.name
+
     def validate(self, log=None, validators=None):
         validators = validators or []
         validators.extend(VALIDATORS)
@@ -488,7 +508,7 @@ class Dataset(object):
 
     @property
     def directory(self):
-        return self.tablegroup._fname.parent
+        return self.tablegroup._fname.parent if self.tablegroup._fname else self.tablegroup.base
 
     @property
     def module(self):
@@ -514,15 +534,18 @@ class Dataset(object):
     #
     @classmethod
     def from_metadata(cls, fname):
-        fname = pathlib.Path(fname)
-        if fname.is_dir():
-            name = '{0}{1}'.format(cls.__name__, MD_SUFFIX)
-            tablegroup = TableGroup.from_file(pkg_path('modules', name))
-            # adapt the path of the metadata file such that paths to tables are resolved
-            # correctly:
-            tablegroup._fname = fname.joinpath(name)
+        if is_url(fname):
+            tablegroup = TableGroup.from_url(fname)
         else:
-            tablegroup = TableGroup.from_file(fname)
+            fname = pathlib.Path(fname)
+            if fname.is_dir():
+                name = '{0}{1}'.format(cls.__name__, MD_SUFFIX)
+                tablegroup = TableGroup.from_file(pkg_path('modules', name))
+                # adapt the path of the metadata file such that paths to tables are resolved
+                # correctly:
+                tablegroup._fname = fname.joinpath(name)
+            else:
+                tablegroup = TableGroup.from_file(fname)
 
         comps = collections.Counter()
         for table in tablegroup.tables:
@@ -681,7 +704,7 @@ class Dataset(object):
                 sum(1 for _ in table) if (exact or 'dc:extent' not in table.common_props)
                 else int(table.common_props.get('dc:extent'))))
         if self.sources:
-            res.append((self.bibpath.name, 'Sources', len(self.sources)))
+            res.append((self.bibname, 'Sources', len(self.sources)))
         return res
 
     def write_metadata(self, fname=None):
@@ -714,6 +737,29 @@ class Dataset(object):
                 item[v] = item[k]
             yield item
 
+    def objects(self, table, cls=None):
+        """
+        Read data of a CLDF component as `pycldf.orm.Object` instances.
+
+        :param table: table to read, specified as component name.
+        :param cls: `orm.Object` subclass to instantiate objects with.
+        :return:
+        """
+        cls = cls or ORM_CLASSES[table]
+
+        # ORM usage is read-only, so we can cache the objects.
+        if table not in self._objects:
+            for item in self[table]:
+                item = cls(self, item)
+                self._objects[table][item.id] = item
+
+        return DictTuple(self._objects[table].values())
+
+    def get_object(self, table, id_, cls=None):
+        if table not in self._objects:
+            self.objects(table, cls=cls)
+        return self._objects[table][id_]
+
 
 class Generic(Dataset):
     @property
